# cluemapper 0.8.4

from pkg_resources import resource_filename
from trac import core as traccore
from trac.admin import api as adminapi
from trac.web import chrome
from trac.util.translation import _
from trac.ticket.notification import TicketNotificationSystem
from trac.notification import NotificationSystem


class CommunicationAdminPanel(traccore.Component):
    traccore.implements(adminapi.IAdminPanelProvider,
                        chrome.ITemplateProvider)

    def render_admin_panel(self, req, cat, page, path_info):
        message = None

        notsys_ticket = TicketNotificationSystem(self.env)
        notsys_general = NotificationSystem(self.env)

        data = {'message': message}

        if req.args.get('communication_applied', False):
            self.config.set('notification', 'always_notify_owner',
                            req.args.get('owner', False) == '1')
            self.config.set('notification', 'always_notify_updater',
                            req.args.get('updater', False) == '1')
            self.config.set('notification', 'always_notify_reporter',
                            req.args.get('reporter', False) == '1')
            # POR
            self.config.set('notification', 'smtp_always_cc',
                            req.args.get('alwayscc', ''))

            self.config.set('notification', 'smtp_always_bcc',
                            req.args.get('alwaysbcc', ''))

            self.config.save()
            notsys_ticket = TicketNotificationSystem(self.env)
            notsys_general = NotificationSystem(self.env)
            message = u'Communication saved'

        if notsys_ticket.always_notify_owner:
            data['owner'] = True
        if notsys_ticket.always_notify_updater:
            data['updater'] = True
        if notsys_ticket.always_notify_reporter:
            data['reporter'] = True
        data['alwayscc'] = notsys_general.smtp_always_cc
        data['alwaysbcc'] = notsys_general.smtp_always_bcc

        return 'admin-communication.html', data

    def get_admin_panels(self, req):
        if req.perm.has_permission('TRAC_ADMIN'):
            yield ('general', _('General'),
                   'communication', 'Communication')

    def get_htdocs_dirs(self):
        return []

    def get_templates_dirs(self):
        return [resource_filename(__name__, 'templates')]



# -*- coding: utf-8 -*-
# inspired by https://github.com/dr4Ke/ldap-to-svn-authz/blob/master/sync_ldap_groups_to_svn_authz.py

from ConfigParser import ConfigParser

from penelope.core.models import DBSession
from penelope.core.models.dashboard import User, Subversion


def repo_authz(repo, authz, users):
    if not repo.svn_name:
        return authz

    section = '%s:/' % repo.svn_name
    if not authz.has_section(section):
        authz.add_section(section)
    for user in users:
        roles = user.roles_in_context(repo.project).copy()
        acl = [(a.role_id, a.permission_name) for a in repo.acl]
        acl.append(('administrator', 'edit'))
        acl.append(('administrator', 'view'))
        permissions = {}
        for x,y in acl:
            if x in roles:
                permissions.setdefault(x, []).append(y)
        def perms(p):
            if 'edit' in p and 'view' in p:
                return 'rw'
            elif 'edit' in p:
                return 'w'
            elif 'view' in p:
                return 'r'
        permissions = [perms(k[1]) for k in permissions.items()]
        # to be sure the first is the longest (rw)
        permissions.sort(lambda x,y: cmp(len(x), len(y)), reverse=True)
        if permissions:
            authz.set(section, user.svn_login, permissions[0])


def generate_authz(settings):
    """
        [groups]
        @admin = haren

        ###
        ### deny all but admins to the tree
        ###

        [/]
        * =
        @admin = rw

        ###
        ### allow more specific people on a per-repo basis below
        ###

        [repo1:/]
        ldap-user1 = rw
        file-user1 = rw

        [repo2:/]
        ldap-user2 = rw
        file-user2 = rw
    """
    svnauth_init = settings.get('por.svn.authz_init')

    db = DBSession()
    authz = ConfigParser()
    authz.read(svnauth_init)

    users = db.query(User).filter_by(active=True).all()
    for repo in db.query(Subversion).all():
        repo_authz(repo, authz, users)

    authz_file = settings.get('por.svn.authz')
    with open(authz_file, 'wb') as configfile:
        authz.write(configfile)

    return authz



# -*- coding: utf-8 -*-

from sqlalchemy import event
from penelope.core.models.dashboard import Application
from penelope.core.models.dashboard import modify_application_type, TRAC, SVN
from penelope.trac.populate import add_trac_to_project, add_svn_to_project


def app_insert_listener(mapper, connection, target):

    modify_application_type(mapper, connection, target)

    if not target.api_uri:

        if target.application_type == TRAC:
            add_trac_to_project(target)

        elif target.application_type == SVN:
            add_svn_to_project(target)


event.listen(Application, 'before_insert', app_insert_listener, propagate=True)



# -*- coding: utf-8 -*-

import redis
import datetime
import json
import operator
import pkg_resources
import re
import urllib
import mandrill
import sqlalchemy.orm.exc

from redis.exceptions import ConnectionError
from json import dumps
from pytz import timezone
from genshi.builder import tag
from genshi.core import TEXT
from genshi.output import TextSerializer
from genshi.filters.transform import Transformer

from creole.rest2html.clean_writer import rest2html
from trac.config import IntOption, Option
from trac.core import Component
from trac.core import implements
from trac.resource import Resource
from trac.ticket import query
from trac.wiki.formatter import format_to_html
from trac.test import Mock, MockPerm
from trac.web.href import Href
from trac.mimeview import Context
from themeengine.api import ThemeBase
from tracrpc.api import IXMLRPCHandler
from trac.ticket.model import Milestone
from trac.ticket.default_workflow import ConfigurableTicketWorkflow
from trac.ticket.api import ITicketChangeListener
from trac.ticket.api import ITicketActionController
from trac.web.api import IRequestFilter, ITemplateStreamFilter, IRequestHandler
from trac.ticket.api import TicketSystem
from trac.web.chrome import ITemplateProvider, add_script, add_script_data, add_stylesheet, Chrome
from trac.ticket.web_ui import TicketModule
from trac.notification import IEmailSender

from penelope.core.fanstatic_resources import dashboard
from penelope.core.fanstatic_resources import add_entry_from_ticket
from penelope.core.models import DBSession
from penelope.core.models.dashboard import CustomerRequest, Project, User
from penelope.core.models.tp import TimeEntry
from penelope.core.models.tp import timedelta_as_human_str
from genshi import Markup

from penelope.trac.i18n import add_domains



def del_script(req, script):
    if not req.chrome.get('scripts'):
        return False
    req.chrome['scripts'] = \
        [s for s in req.chrome['scripts'] \
            if not s.get('href', '').endswith(script)]


def del_style(req, script):
    all_styles = req.chrome['links']['stylesheet']
    new_styles = [
            s for s in all_styles
            if not s.get('href', '').endswith(script)
            ]
    if new_styles != all_styles:
        req.chrome['links']['stylesheet'] = new_styles
        return True
    return False


class PorTheme(ThemeBase):
    """A theme for Trac"""
    template = htdocs = css = True

    def __init__(self):
        add_domains(self.env.path)



class PorNav(Component):
    """ add custom breadcrumbs """
    implements(ITemplateStreamFilter)


    # ITemplateStreamFilter methods
    # TODO: cache
    # TODO: c'è un metodo per ricavare le url del customer e del project?
    def filter_stream(self, req, method, filename, stream, data):
        project_id = self.env.config.get('por-dashboard', 'project-id')
        if project_id:
            project = DBSession().query(Project).get(project_id)
            # XXX se project is None, 404

            stream |= Transformer(".//div[@id='trac-before-subnav']").prepend(tag.ul(
                    tag.li(tag.a("Home", href="/")),
                    tag.li(
                            tag.span(" / ", class_="divider"),
                            tag.a(project.customer.name, href="/admin/Customer/%s" % project.customer.id)
                        ),
                    tag.li(
                            tag.span(" / ", class_="divider"),
                            tag.a(project.name, href="/admin/Project/%s" % project.id)
                        ),
                    tag.li(
                            tag.span(" / ", class_="divider"),
                            tag.a('Trac', href="/trac/%s" % project.id),
                            class_='active'
                        ),
                    class_="breadcrumb noprint",
                ))

        return stream



class PorFanstatic(Component):
    """ """
    implements(IRequestFilter, ITemplateProvider)

    # ITemplateProvider methods
    def get_htdocs_dirs(self):
        return [('penelope', pkg_resources.resource_filename(__name__, 'htdocs'))]

    def get_templates_dirs(self):
        return [pkg_resources.resource_filename(__name__, 'templates')]

    # IRequestFilter methods
    def pre_process_request(self, req, handler):
        return handler

    def post_process_request(self, req, template, data, content_type):
        # remove script (jquery aggiornato caricato via fanstatic)
        del_script(req, 'common/js/jquery.js')
        # fanstatic
        dashboard.need()
        add_script(req, 'penelope/por.js')
        add_stylesheet(req, 'penelope/por.css')
        field_descriptions = dict([(field['name'], field['descriptions']) \
            for field in TicketSystem(self.env).get_custom_fields() \
                if field.get('descriptions')])
        add_script_data(req, {'fielddescriptions': field_descriptions})
        return template, data, content_type



class RescopedCSS(Component):
    """ """
    implements(IRequestFilter)

    def pre_process_request(self, req, handler):
        return handler

    def post_process_request(self, req, template, data, content_type):
        # TODO see what's available at runtime
        for cssfile in ['about.css', 'admin.css', 'browser.css', 'changeset.css', 'code.css',
                        'diff.css', 'prefs.css', 'report.css', 'roadmap.css', 'search.css',
                        'ticket.css', 'timeline.css', 'trac.css', 'wiki.css']:
            if del_style(req, 'common/css/%s' % cssfile):
                add_stylesheet(req, 'penelope/rescoped-css/%s' % cssfile)
        return template, data, content_type



class PorTimeEntry(Component):
    """
    Adds the required javascript for the feature "Add Time Entry from Ticket".
    """
    implements(IRequestFilter)

    # IRequestFilter methods
    def pre_process_request(self, req, handler):
        return handler

    def post_process_request(self, req, template, data, content_type):
        if template == 'ticket.html' and req.perm.has_permission('TIME_ENTRY_ADD'):
            cr = DBSession().query(CustomerRequest).get(data['ticket'].values['customerrequest'])
            if cr and cr.workflow_state in ['created', 'estimated']:
                add_entry_from_ticket.need()
        return template, data, content_type


class PorModifySimple(Component):
    """
    Hides (not remove) some fields from the Modify Ticket if the user is a customer.
    """
    implements(IRequestFilter)

    # IRequestFilter methods
    def pre_process_request(self, req, handler):

        if req.perm.has_permission('SENSITIVE_VIEW'):       # XXX right permission here
            hidden_cls = 'hidden-to-customers'
        else:
            hidden_cls = 'hide'

        field_classes = [
            ('type', ''),
            ('priority', ''),
            ('customerrequest', ''),
            ('cc', ''),
            ('keywords', ''),
            ('milestone', ''),
            ('fasesviluppo', ''),
            ('issuetype', ''),
            ('esogeno', hidden_cls),
            ('stats_exclude', hidden_cls),
            ('version', hidden_cls),
            ('component', hidden_cls),
            ('blocking', hidden_cls),
            ('blockedby', hidden_cls),
            ('sensitive', hidden_cls)
            ]

        field_order = [f[0] for f in field_classes]

        def sortkey(field):
            try:
                return field_order.index(field['name'])
            except ValueError:
                return len(field_order) + 1000

        req.ticketfields_sortkey = sortkey
        req.ticketfields_classes = dict(field_classes)

        return handler

    def post_process_request(self, req, template, data, content_type):
        return template, data, content_type


class PorTicketTimeEntries(Component):
    """
    Render ticket timeentries
    """
    implements(IRequestFilter)

    # IRequestFilter methods
    def pre_process_request(self, req, handler):
        if isinstance(handler, Chrome):
            # return ASAP if we are serving a static resource
            return handler

        req.ticket_time_entries = []
        if isinstance(handler, TicketModule):
            project_id = self.env.config.get('por-dashboard', 'project-id')
            ticket_id = req.args.get('id', None)
            if ticket_id and project_id:
                timeentries = DBSession().query(TimeEntry)\
                               .filter_by(project_id=project_id)\
                               .filter_by(ticket=int(ticket_id)).order_by(TimeEntry.date.desc(), TimeEntry.modification_date.desc())
                delta = datetime.timedelta()
                delta_tot = sum([e.hours for e in timeentries], delta)
                req.ticket_time_entries_total = timedelta_as_human_str(delta_tot)
                req.ticket_time_entries = timeentries[:20]
                req.ticket_time_entries_count = timeentries.count()

        return handler

    def post_process_request(self, req, template, data, content_type):
        return template, data, content_type


class PorReportDropDown(Component):
    """
    Render report list as a bootstrap button/dropdown combo
    """
    implements(IRequestFilter)

    def _iter_reports(self):
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("SELECT id, title FROM report ORDER BY id")
        for row in cursor:
            yield row[0], row[1]
            if row[0] in (3, 6):
                yield None, None

    def genshi_text(self, stream, flag=False):
        """
        @author: mmariani, lucabel
        @description: we use this method in template 'cause we can't
        draw directly the elem from chrome.ctxtnav: it put a link in
        a button and we don't want that
        """
        return Markup(' '.join(TextSerializer()(stream)))

    # IRequestFilter methods
    def pre_process_request(self, req, handler):
        if isinstance(handler, Chrome):
            # return ASAP if we are serving a static resource
            return handler

        lis = []
        for report_id, report_name in self._iter_reports():
            lis.append({'name': report_name, 'href': req.href.report(report_id)})

        req.report_dropdown_elements = lis
        req.genshi_text = self.genshi_text

        return handler

    def post_process_request(self, req, template, data, content_type):
        return template, data, content_type


class PorUserEmailLookup(Component):
    """
    Replaces email in workflow actions with a fullname
    """
    implements(ITemplateStreamFilter)

    def email_lookup(self, text):
        try:
            user = DBSession.query(User).filter(User.email==text).one()
            return user.fullname
        except sqlalchemy.orm.exc.NoResultFound:
            return text

    def filter_stream(self, req, method, filename, stream, data):
        stream |= Transformer("//select[@id='action_review_reassign_owner']/option").map(self.email_lookup, TEXT)
        stream |= Transformer("//select[@id='action_reassign_reassign_owner']/option").map(self.email_lookup, TEXT)
        stream |= Transformer("//select[@name='0_owner']/option").map(self.email_lookup, TEXT)
        stream |= Transformer("//select[@name='field_owner']/option").map(self.email_lookup, TEXT)
        return stream


class TicketRPC(Component):
    """ An interface to Trac's ticketing system. """

    implements(IXMLRPCHandler)

    # IXMLRPCHandler methods
    def xmlrpc_namespace(self):
        return 'ticket'

    def xmlrpc_methods(self):
        yield (None, ((list,), (list, str)), self.queryWithDetails)
        yield (None, ((list,), (list, str)), self.queryCustomerRequestsByTicktes)
        yield (None, ((list,), (list, str)), self.queryAllCustomerRequests)

    # Exported methods
    def queryWithDetails(self, req, qstr='status!=closed'):
        """
        Perform a ticket query, returning a list of ticket dictionaries.
        All queries will use stored settings for maximum number of results per
        page and paging options. Use `max=n` to define number of results to
        receive, and use `page=n` to page through larger result sets. Using
        `max=0` will turn off paging and return all results.
        """
        # TODO: gestione custom id^= (workaround)
        if "id^=" in qstr:
            startswith = re.search('id\^=([\d]*)', qstr).group(1)
            qstr = re.sub('id\^=[\d]*', '', qstr)
            qstr = re.sub('^&|&$|&&', '', qstr)
            if "max=" in qstr:
                limit = re.search('max=([\d]*)', qstr).group(1)
                qstr = re.sub('max=[\d]*', 'max=0', qstr)
            else:
                limit = None
        else:
            startswith = None
            limit = None

        q = query.Query.from_string(self.env, qstr)
        ticket_realm = Resource('ticket')
        out = []
        for t in q.execute(req):
            tid = t['id']
            if not startswith or str(tid).startswith(startswith):
                if 'TICKET_VIEW' in req.perm(ticket_realm(id=tid)):
                    out.append(t)
                    if limit and len(out) >= limit:
                        break

        # fill 'resolution' value (it is not provided by the above query)
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("SELECT id, resolution FROM ticket WHERE status='closed'")

        resolution = {}
        for row in cursor:
            resolution[row[0]] = row[1]

        # fill 'sensitive' value (TODO optimize, as it now retrieves all tickets)
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("SELECT ticket, value FROM ticket_custom WHERE name='sensitive' AND value='1'")
        sensitive = set()
        for row in cursor:
            sensitive.add(row[0])

        # fill 'customerrequest' value (TODO optimize, as it now retrieves all tickets)
        cursor = db.cursor()
        cursor.execute("SELECT t.id AS ticket, c.value FROM ticket t INNER JOIN ticket_custom c ON (t.id = c.ticket AND c.name = 'customerrequest') order by ticket")
        cr= {}
        for row in cursor:
            cr[row[0]]=row[1]

        for t in out:
            t['sensitive'] = int(t['id'] in sensitive)
            t['resolution'] = resolution.get(t['id'], '')
            t['cr'] = cr.get(t['id'], '')

        return out

    def queryCustomerRequestsByTicktes(self, req, ticket_ids):
        """
            Args:
                ticket_ids: list of ticket ids

            Returns:
                list of tuples: ticket_id, customerrequest_field ('id> description')
        """
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("""SELECT ticket, value FROM ticket_custom
            WHERE name='customerrequest' AND ticket IN %(tickets)s;""",
                {'tickets': tuple(ticket_ids)})
        customer_request_ids = cursor.fetchall() or []
        db.rollback()
        return customer_request_ids

    def queryAllCustomerRequests(self, req):
        """
            Args:
                ticket_ids: list of ticket ids

            Returns:
                list of tuples: ticket_id, customerrequest_field
        """
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("SELECT ticket, value FROM ticket_custom WHERE name='customerrequest'")
        rows = cursor.fetchall() or []
        db.rollback()
        return rows


class CustomerTicketsPolicy(Component):
    implements(ITemplateStreamFilter)

    # ITemplateStreamFilter methods
    def filter_stream(self, req, method, filename, stream, data):
        if filename == 'ticket.html':
            ticket = data['ticket']
            if ticket.exists: # only for new tickets
                return stream
            if req.perm.has_permission('SENSITIVE_VIEW'):
                # probably there is a better way to check the customer
                # but right now it's the only reasonable
                return stream
            stream |= Transformer("//input[@id='field-esogeno']").attr('checked', 'checked')
        return stream


class InactiveCRTicketsPolicy(Component):
    implements(ITemplateStreamFilter)

    # ITemplateStreamFilter methods
    def filter_stream(self, req, method, filename, stream, data):
        """Return a filtered Genshi event stream, or the original unfiltered
        stream if no match.

        `req` is the current request object, `method` is the Genshi render
        method (xml, xhtml or text), `filename` is the filename of the template
        to be rendered, `stream` is the event stream and `data` is the data for
        the current template.

        See the Genshi documentation for more information.
        """
        if filename == 'ticket.html':
            ticket = data['ticket']
            if ticket.exists:
                if req.perm.has_permission('SENSITIVE_VIEW'):
                    qry = DBSession().query(CustomerRequest)
                    if not qry.get(ticket.values.get('customerrequest')).active:
                        div = tag.div(
                            tag.div(
                                tag.strong(u'Heads up! '),
                                tag.span(u'This ticket is assigned to an inactive customer request.',),
                                class_="alert alert-info"),
                            id='inactive_cr_ticket')
                        return stream | Transformer("//div[@id='ticket']").before(div)
        return stream


class SensitiveTicketsPolicy(Component):
    implements(ITemplateStreamFilter)

    # ITemplateStreamFilter methods
    def filter_stream(self, req, method, filename, stream, data):
        """Return a filtered Genshi event stream, or the original unfiltered
        stream if no match.

        `req` is the current request object, `method` is the Genshi render
        method (xml, xhtml or text), `filename` is the filename of the template
        to be rendered, `stream` is the event stream and `data` is the data for
        the current template.

        See the Genshi documentation for more information.
        """
        if filename == 'ticket.html':
            ticket = data['ticket']
            if ticket.exists:
                sensitive = (ticket.values.get('sensitive') == '1')
                if req.perm.has_permission('SENSITIVE_VIEW'):
                    action = data['ticket'].exists and \
                        req.href.ticket(data['ticket'].id) or \
                        req.href.newticket()
                    form = tag.form(
                        tag.input(
                            id="field-sensitive",
                            name="field_sensitive",
                            value=sensitive and '0' or '1',
                            type='hidden'),
                        tag.input(
                            type="hidden",
                            name="ts",
                            value=data['timestamp']),
                        tag.input(
                            type="hidden",
                            name="action",
                            value=data['action']),
                        tag.input(
                            type="submit",
                            name="submit",
                            class_='btn',
                            # TODO: translations
                            value=sensitive and 'make public' or 'make private'),
                        method="post",
                        id="sensitiveform",
                        action="%s#trac-modify-sensitive" % action,
                    )
                    if sensitive:
                        div = tag.div(
                            # TODO: translations
                            tag.div(
                                form,
                                # XXX i18n TODO tag.span(_("trac_por_ticket_sensitive"),
                                tag.span(u'Il ticket è riservato ai developer e non visibile al cliente',
                                         class_="sensitive-message"),
                                class_="alert alert-success"),
                            id='sensitiveticket')
                    else:
                        div = tag.div(
                            # TODO: translations
                            tag.div(
                                form,
                                # XXX i18n TODO tag.span(_("trac_por_ticket_public"),
                                tag.span(u'Ticket condiviso con il cliente',
                                         class_="sensitive-message"),
                                class_="alert alert-error"),
                            id='sensitiveticket')
                    return stream | \
                        Transformer("//div[@id='ticket']").before(div)
        return stream




class OutstandingTickets(Component):
    """
    Returns a JSON object with info about open tickets.
    """
    implements(IRequestHandler)

    def match_request(self, req):
        match = re.match(r'/outstanding_tickets/', req.path_info)
        if match:
            return True


    def iter_tickets(self, req, query_string):
        q = query.Query.from_string(self.env, query_string)
        ticket_realm = Resource('ticket')
        for t in sorted(q.execute(req), key=operator.itemgetter('changetime'), reverse=True):
            if 'TICKET_VIEW' in req.perm(ticket_realm(id=t['id'])):
                yield {
                    'id': t['id'],
                    'summary': t['summary'],
                    'status': t['status'],
                    'priority_value': t['priority_value'],
                    'href': t['href'],
                    }

    def process_request(self, req):
        qs = (
            'owner=%s'
            '&max=0'
            '&status=accepted'
            '&status=assigned'
            '&status=new'
            '&status=reopened'
            '&status=reviewing'
            )

        qs_own = qs % req.authname
        qs_unassigned = qs % ''

        tickets_own = list(self.iter_tickets(req, query_string=qs_own))
        tickets_unassigned = list(self.iter_tickets(req, query_string=qs_unassigned))

        data = json.dumps({
                    'tickets_own': tickets_own,
                    'query_url_own': '%s/query?%s' % (req.base_url, qs_own),
                    'tickets_unassigned': tickets_unassigned,
                    'query_url_unassigned': '%s/query?%s' % (req.base_url, qs_unassigned),
                    })
        req.send(data, 'application/json')



class CurrentIteration(Component):
    """
    Redirects to a precompiled custom query for active customer requests.
    """
    implements(IRequestHandler)

    def match_request(self, req):
        match = re.match(r'/current_iteration$', req.path_info)
        if match:
            return True


    def process_request(self, req):

        project_id = self.env.config.get('por-dashboard', 'project-id')

        qry = DBSession().query(CustomerRequest)
        qry = qry.filter(CustomerRequest.project_id==project_id)
        qry = qry.filter(CustomerRequest.workflow_state.in_(['created', 'estimated']))

        query_params = {
                'owner': ['$USER', ''],
                'status': ['assigned', 'new', 'reopened', 'reviewing'],
                'group': 'customerrequest',
                'customerrequest': [ cr.id for cr in qry ],
                'col': ['id', 'summary', 'status', 'type', 'priority', 'component', 'customerrequest'],
                'order': 'priority',
                }

        query_string = urllib.urlencode(query_params, doseq=True)

        req.redirect('%s/query?%s' % (req.base_url, query_string))


class MilestoneEnhacement(Component):
    """
    Add milestone due date to selection
    """
    implements(ITemplateStreamFilter)

    def duedate_lookup(self, text):
        milestone = Milestone(self.env, text)
        if milestone.due:
            tzname = self.env.config.get('trac', 'default_timezone')
            if tzname:
                tz = timezone(tzname)
                due = milestone.due.astimezone(tz)
            else:
                due = milestone.due
            text += ' [%s]' % due.strftime('%Y-%m-%d')
        return text

    def filter_stream(self, req, method, filename, stream, data):
        stream |= Transformer("//select[@id='field-milestone']/optgroup/option").map(self.duedate_lookup, TEXT)
        stream |= Transformer("//a[@class='milestone']").map(self.duedate_lookup, TEXT)
        return stream



class MandrillEmailSender(Component):
    implements(IEmailSender)

    smtp_server = Option('notification', 'smtp_host', 'localhost',
        """SMTP server hostname to use for email notifications.""")

    smtp_port = IntOption('notification', 'smtp_port', 25,
        """SMTP server port to use for email notification.""")

    smtp_user = Option('notification', 'smtp_username', '',
        """Username for SMTP server. (''since 0.9'')""")

    smtp_password = Option('notification', 'smtp_password', '',
        """Password for SMTP server. (''since 0.9'')""")

    def wiki2html(self, wiki):
        """ The easiest way to convert wiki to html """
        req = Mock(href=Href(self.env.abs_href.base),
                   abs_href=self.env.abs_href,
                   authname='anonymous',
                   perm=MockPerm(),
                   args={})
        context = Context.from_request(req, 'wiki')
        try:
            html = format_to_html(self.env, context, wiki).encode('utf8','ignore')
        except AttributeError:
            html = wiki
        return html

    def send(self, from_addr, recipients, data):
        # Ensure the message complies with RFC2822: use CRLF line endings
        message = data['msg']
        data = data['data']

        params = {}
        changes_body = data['changes_body']
        if changes_body:
            params['changes_body'] = rest2html(changes_body)

        if data['ticket']['new']:
            params['ticket_new'] = True

        params['ticket_body_hdr'] = data['ticket_body_hdr']
        params['ticket_link'] = data['ticket']['link']
        params['ticket_reporter'] = data['ticket']['reporter']
        params['ticket_owner'] = data['ticket']['owner']
        params['ticket_description'] = self.wiki2html(data['ticket']['description'])
        params['ticket_type'] = data['ticket']['type']
        params['ticket_status'] = data['ticket']['status']
        params['ticket_priority'] = data['ticket']['priority']
        params['ticket_milestone'] = data['ticket'].get('milestone')
        params['change_author'] = data['change'].get('author','')
        params['change_comment'] = self.wiki2html(data['change'].get('comment',''))
        params['project_name'] = data['project']['name']
        params['project_url'] = data['project']['url']

        merged_params = []
        for k,v in params.items():
            merged_params.append({'name': k, 'content':v})

        mandrill_client = mandrill.Mandrill(self.smtp_password)
        message = {'auto_html': None,
                   'auto_text': None,
                   'from_email': from_addr,
                   'from_name': 'RedTurtle Team',
                   'headers': {'Reply-To': from_addr},
                   'important': True,
                   'inline_css': True,
                   'global_merge_vars': merged_params,
                   'subject': unicode(message['subject']),
                   'to': [],
                   }
        for rec in recipients:
            message['to'].append({'email':rec})

        self.log.info("Sending notification through Mandril API to %s"
                      % recipients)

        mandrill_client.messages.send_template(template_name='ticket',
                                               template_content=[],
                                               message=message)


class TicketWorkflowOpBase(Component):
    """Abstract base class for 'simple' ticket workflow operations."""

    implements(ITicketActionController)
    abstract = True

    _op_name = None # Must be specified.

    def get_configurable_workflow(self):
        controllers = TicketSystem(self.env).action_controllers
        for controller in controllers:
            if isinstance(controller, ConfigurableTicketWorkflow):
                return controller
        return ConfigurableTicketWorkflow(self.env)

    # ITicketActionController methods

    def get_ticket_actions(self, req, ticket):
        """Finds the actions that use this operation"""
        controller = self.get_configurable_workflow()
        return controller.get_actions_by_operation_for_req(req, ticket,
                                                           self._op_name)

    def get_all_status(self):
        """Provide any additional status values"""
        # We don't have anything special here; the statuses will be recognized
        # by the default controller.
        return []

    # This should most likely be overridden to be more functional
    def render_ticket_action_control(self, req, ticket, action):
        """Returns the action control"""
        actions = self.get_configurable_workflow().actions
        label = actions[action]['name']
        return (label, tag(''), '')

    def get_ticket_changes(self, req, ticket, action):
        """Must be implemented in subclasses"""
        raise NotImplementedError

    def apply_action_side_effects(self, req, ticket, action):
        """No side effects"""
        pass


class TicketWorkflowOpOwnerPrevious(TicketWorkflowOpBase):
    """Sets the owner to the previous owner

    Don't forget to add the `TicketWorkflowOpOwnerPrevious` to the workflow
    option in [ticket].
    If there is no workflow option, the line will look like this:

    workflow = ConfigurableTicketWorkflow,TicketWorkflowOpOwnerPrevious
    """

    _op_name = 'set_owner_to_previous'

    # ITicketActionController methods

    def render_ticket_action_control(self, req, ticket, action):
        """Returns the action control"""
        actions = self.get_configurable_workflow().actions
        label = actions[action]['name']
        new_owner = self._new_owner(ticket, req.authname)
        if new_owner:
            hint = 'The owner will change to %s' % new_owner
        else:
            hint = 'The owner will be deleted.'
        control = tag('')
        return (label, control, hint)

    def get_ticket_changes(self, req, ticket, action):
        """Returns the change of owner."""
        return {'owner': self._new_owner(ticket, req.authname)}

    def _new_owner(self, ticket, authname):
        """Determines the new owner"""
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("SELECT oldvalue FROM ticket_change WHERE ticket=%s " \
                       "AND field='owner' AND oldvalue != '' ORDER BY -time", (ticket.id, ))
        row = cursor.fetchone()
        if row:
            owner = row[0]
        else: # The owner has never changed.
            owner = ticket['reporter']
        if owner == authname:
            owner = ticket['reporter']
        return owner


class TicketChangePublisher(Component):
    implements(ITicketChangeListener)

    def __init__(self, *args, **kwargs):
        pass

    def notify_redis(self, ticket):
        r = redis.StrictRedis()
        value = dict([(a, ticket.values.get(a)) for a in ('summary', 'owner', 'customerrequest', 'priority')])
        id_ = '{trac_id}#{ticket_id}'.format(trac_id=self.env.config['por-dashboard'].get('project-id'),
                                             ticket_id=ticket.id)
        try:
            r.publish('*', dumps({'event': 'ticket_changed', 'data': {id_: value}}))
        except ConnectionError:
            pass

    def notify_feedly(self, ticket, comment, author):
        from penelope.core.activity_stream import add_activity

        users_to_notify = set()
        users_to_notify.add(ticket.values.get('reporter'))
        users_to_notify.add(ticket.values.get('owner'))
        try:
            users_to_notify.remove(author)
        except KeyError:
            pass

        absolute_path = '/trac/{trac_id}/ticket/{ticket_id}'.format(ticket_id=ticket.id, trac_id=self.env.config['por-dashboard'].get('project-id'))
        message = "Ticket #{ticket} has been {comment} in {trac_name}.".format(ticket=ticket.id, comment=comment, trac_name=self.config['project'].get('name'))
        created_by = author
        user_ids = []

        for email in users_to_notify:
            try:
                user_ids.append(DBSession.query(User.id).filter(User.email==email).one().id)
            except sqlalchemy.orm.exc.NoResultFound:
                pass
        try:
            add_activity(user_ids, message, absolute_path, created_by)
        except ConnectionError:
            pass

    def realms(self):
        yield 'ticket'

    def ticket_created(self, ticket):
        self.notify_feedly(ticket, 'created', ticket.values.get('reporter'))

    def ticket_changed(self, ticket, comment, author, old_values):
        self.notify_redis(ticket)
        self.notify_feedly(ticket, 'changed', author)

    def ticket_deleted(self, ticket):
        pass



# -*- coding: utf-8 -*-

from trac.util.datefmt import utc
from trac.ticket.model import Ticket
from trac.ticket.notification import TicketNotifyEmail
from trac.test import EnvironmentStub, Mock, MockPerm
from trac.tests.notification import SMTPThreadedServer
from trac.tests.notification import smtp_address

import os
import unittest

SMTP_TEST_PORT = 7000 + os.getpid() % 1000
MAXBODYWIDTH = 76



class NotificationTestCase(unittest.TestCase):
    """Notification test cases that send email over SMTP"""

    def setUp(self):
        self.env = EnvironmentStub(default_data=True)
        self.env.config.set('project', 'name', 'TracTest')
        self.env.config.set('notification', 'smtp_enabled', 'true')
        self.env.config.set('notification', 'always_notify_owner', 'true')
        self.env.config.set('notification', 'always_notify_reporter', 'true')
        self.env.config.set('notification', 'smtp_always_cc',
                                                'frank.developer@example.org')
        # BBB: mark.customer@example.net chi è in always_cc viene notificato
        # sempre anche se non ha il permesso SENSITIVE_VIEW
        # è corretto?
        self.env.config.set('notification', 'use_public_cc', 'true')
        self.env.config.set('notification', 'smtp_port', str(SMTP_TEST_PORT))
        self.env.config.set('notification', 'smtp_server','localhost')
        self.req = Mock(href=self.env.href, abs_href=self.env.abs_href, tz=utc,
                        perm=MockPerm())
        self.smtpd = SMTPThreadedServer(SMTP_TEST_PORT)
        self.smtpd.start()

    def tearDown(self):
        """Signal the notification test suite that a test is over"""
        self.smtpd.cleanup()
        self.smtpd.stop()
        self.env.reset_db()

    def test_recipients(self):
        """To/Cc recipients"""
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.executemany("INSERT INTO permission VALUES (%s,%s)", [
                           ('frank.developer@example.org', 'SENSITIVE_VIEW'), 
                           ('joe.developer@example.org', 'SENSITIVE_VIEW'), 
                           ('jim.developer@example.org', 'SENSITIVE_VIEW'), 
                           ])
        ticket = Ticket(self.env)
        ticket['reporter'] = 'mary.customer@example.org'
        ticket['owner']    = 'joe.developer@example.org'
        ticket['cc']       = 'jim.developer@example.org, joe.customer@example.net, ' \
                             'jack.customer@example.net'
        ticket['summary'] = 'Foo'
        ticket['sensitive'] = '1'
        ticket.insert()
        tn = TicketNotifyEmail(self.env)
        tn.notify(ticket, newticket=True)
        recipients = self.smtpd.get_recipients()
        # checks there is no duplicate in the recipient list
        rcpts = []
        for r in recipients:
            self.failIf(r in rcpts)
            rcpts.append(r)
        # checks that all cc recipients have been notified
        cc_list = self.env.config.get('notification', 'smtp_always_cc')
        cc_list = "%s, %s" % (cc_list, ticket['cc'])
        for r in cc_list.replace(',', ' ').split():
            if 'customer' in r:
                self.failIf(r in recipients)
            else:
                self.failIf(r not in recipients)
        # checks that owner has not been notified
        self.failIf(smtp_address(ticket['owner']) not in recipients)
        # checks that reporter has not been notified (is a customer)
        self.failIf(smtp_address(ticket['reporter']) in recipients)



# -*- coding: utf-8 -*-
import pkg_resources

from trac.util.translation import domain_functions


_, tag_, N_, add_tracpor_domain = \
            domain_functions('penelope.trac', ('_', 'tag_', 'N_', 'add_domain'))


def add_domains(env_path):
    locale_dir = pkg_resources.resource_filename(__name__, 'locale')
    add_tracpor_domain(env_path, locale_dir)



# -*- coding: utf-8 -*-
import os
import stat
import lipsum
import random
import subprocess
from datetime import date

from plone.i18n.normalizer import idnormalizer
from pyramid.threadlocal import get_current_registry
from trac.env import Environment
from trac.ticket.model import Ticket
from trac.admin.console import run

from penelope.core.models.dbsession import DBSession


def get_reports(**kwargs):
    return (
             ("All Tickets, Active Customer Request",
              """* List all tickets, active customer request.""",
              """SELECT
                 t.id AS ticket
                 , t.summary
                 , t.component
                 , t.version
                 , t.severity
                 , t.milestone
                 , t.status
                 , t.owner
                 , t.reporter
                 , t.time AS created
                 , t.changetime AS modified  -- ## Dates are formatted
                 , c.value AS customerrequest
                 -- , t.description AS _description_  -- ## Uses a full row
              FROM ticket t
              INNER JOIN ticket_custom c ON (t.id = c.ticket AND c.name = 'customerrequest')
              INNER JOIN public.customer_requests cr ON c.value = cr.id AND cr.project_id = '%(project_id)s'
              WHERE cr.placement = 2
              """ % kwargs,),
    )


def _execute(cmd):
        try:
            ret = subprocess.check_call(cmd, stderr=subprocess.STDOUT)
        except OSError, e:
            raise Exception('The cmd %r could not be executed, %r'
                            'full cmd args are: %r'
                                % (cmd[0], e, ' '.join(cmd)))
        return ret


POST_COMMIT_HOOK = """#!/bin/sh
REPOS="$1"
REV="$2"
%(trac_admin)s %(trac_path)s changeset added $REPOS $REV
"""


def add_svn_to_project(application):
    from penelope.core.models.dashboard import Project

    project = DBSession.query(Project).get(application.project_id)
    settings = get_current_registry().settings

    svnenvs = settings.get('penelope.svn.envs')
    tracenvs = settings.get('penelope.trac.envs')
    trac_admin = os.path.abspath('%s/../../bin/trac-admin' % tracenvs)

    if not os.path.exists(svnenvs):
        os.mkdir(svnenvs)

    svnname = application.svn_name
    if svnname:
        svn_path = '%s/%s' % (svnenvs, svnname)
        if not os.path.exists(svn_path):
            raise OSError, 'The path %s doesn\'t exists!' % svn_path
    else:
        idx = ''
        while (not svnname):
            svnname = idnormalizer.normalize("%s%s" % (project.id, idx))
            svn_path = '%s/%s' % (svnenvs, svnname)
            if os.path.exists(svn_path):
                idx = idx and (idx+1) or 1
                svnname = None

        _execute(['svnadmin', '--config-dir', '/etc/subversion', 'create', svn_path])

    for trac in project.tracs:
        tracname = str(trac.trac_name)
        trac_path = '%s/%s' % (tracenvs, tracname)
        run([trac_path, 'repository add %s %s' % (svnname, svn_path)])
        run([trac_path, 'repository resync %s' % (svnname)])
        run([trac_path, 'config set trac repository_sync_per_request ""'])

        hook_filename = '%s/hooks/post-commit' % svn_path
        with open(hook_filename, 'w') as hook:
            opts = {'trac_path': trac_path, 'trac_admin': trac_admin}
            hook.write(POST_COMMIT_HOOK % opts)
        st = os.stat(hook_filename)
        os.chmod(hook_filename, st.st_mode | stat.S_IEXEC)

    application.api_uri = 'svn://%s' % svnname
    application.svn_name = svnname


def add_trac_to_project(application,
        smtp_enabled=True,
        privatecomments=True,
        sensitivetickets=True,
        batchmod=True,
        autocompleteusers=True,
        customfieldadmin=True,
        qafields=True,
        privatetickets=False,
        tracwysiwyg=True,
        attachment_max_size=10485760,
        milestones=[],
        tickets=[],
        project_name=u'',
        ):

    from penelope.core.models.dashboard import Project

    project = DBSession.query(Project).get(application.project_id)

    settings = get_current_registry().settings or application.settings
    tracenvs = settings.get('penelope.trac.envs')

    if not os.path.exists(tracenvs):
        # TODO: logging bootstrap
        os.mkdir(tracenvs)

    tracname = None
    idx = ''
    while (not tracname):
        tracname = idnormalizer.normalize("%s%s" % (project.id, idx))
        trac_path = '%s/%s' % (tracenvs, tracname)
        if os.path.exists(trac_path):
            idx = idx and (idx+1) or 1
            tracname = None

    trac_perm = {
        'administrator': ['TRAC_ADMIN', 'EXTRA_TIMEENTRY'],
        'customer': ['TICKET_CREATE', 'TICKET_MODIFY', 'TICKET_VIEW',
            'WIKI_VIEW', 'XML_RPC',
            'CHANGESET_VIEW', 'FILE_VIEW',
            'LOG_VIEW', 'MILESTONE_VIEW',
            'REPORT_VIEW', 'REPORT_SQL_VIEW',
            'ROADMAP_VIEW', 'SEARCH_VIEW', 'TIMELINE_VIEW'],
        'developer': ['TICKET_CREATE', 'TICKET_MODIFY', 'TICKET_VIEW',
            'WIKI_VIEW', 'XML_RPC',
            'WIKI_CREATE', 'WIKI_MODIFY',
            'CHANGESET_VIEW', 'FILE_VIEW',
            'LOG_VIEW', 'MILESTONE_VIEW',
            'REPORT_VIEW', 'REPORT_SQL_VIEW',
            'ROADMAP_VIEW', 'SEARCH_VIEW',
            'TIMELINE_VIEW', 'REPORT_ADMIN'],
        'internal_developer': ['developer', 'TRAC_ADMIN', 'TIME_ENTRY_ADD'],
        'external_developer': ['developer'],
        'project_manager': ['administrator', 'TIME_ENTRY_ADD'],
    }


    pordb = str(DBSession.bind.url)
    run([trac_path, 'initenv "%s" "%s?schema=trac_%s"' % (
         tracname,
         pordb.replace('postgresql://', 'postgres://'),
         tracname)])

    tracenv = Environment(trac_path)

    # REPORTS
    cnx = tracenv.get_db_cnx().cnx
    cursor = cnx.cursor()
    cursor.executemany(\
        "INSERT INTO report (title, description, query) VALUES (%s, %s, %s)",
        get_reports(project_id=project.id))
    cursor.close()
    cnx.commit()

    # remove report 2
    cursor = cnx.cursor()
    cursor.execute("DELETE FROM report where id=2;")
    cursor.close()
    cnx.commit()

    # update reports
    cursor = cnx.cursor()
    cursor.execute("""UPDATE report set query='
        SELECT p.value AS __color__,
            t.id AS ticket, summary, t.type AS type, cr.name AS CR,
            owner, status,
            time AS created,
            changetime AS _changetime, t.description AS _description,
            reporter AS _reporter
        FROM ticket t
        LEFT JOIN enum p ON p.name = t.priority AND p.type = ''priority''
        LEFT JOIN ticket_custom tc ON t.id = tc.ticket
        LEFT JOIN public.customer_requests cr ON tc.value = cr.id
        WHERE status <> ''closed''
        AND tc.name = ''customerrequest''
        ORDER BY CAST(p.value AS int), milestone, t.type, time' where id=1;""")
    cursor.execute("""UPDATE report set query='
            SELECT p.value AS __color__,
               ''Milestone ''||milestone AS __group__,
               t.id AS ticket, summary, t.type AS type,
               cr.name AS CR,
               owner, status,
               time AS created,
               changetime AS _changetime, t.description AS _description,
               reporter AS _reporter
            FROM ticket t
            LEFT JOIN enum p ON p.name = t.priority AND p.type = ''priority''
            LEFT JOIN ticket_custom tc ON t.id = tc.ticket
            LEFT JOIN public.customer_requests cr ON tc.value = cr.id
            WHERE status <> ''closed''
            AND tc.name = ''customerrequest''
            ORDER BY (milestone IS NULL),milestone, CAST(p.value AS int), t.type, time' where id=3;""")
    cursor.execute("""UPDATE report set query='
            SELECT p.value AS __color__,
               owner AS __group__,
               t.id AS ticket, summary, milestone,  cr.name AS CR, t.type AS type, time AS created,
               changetime AS _changetime, t.description AS _description,
               reporter AS _reporter
            FROM ticket t
            LEFT JOIN enum p ON p.name = t.priority AND p.type = ''priority''
            LEFT JOIN ticket_custom tc ON t.id = tc.ticket
            LEFT JOIN public.customer_requests cr ON tc.value = cr.id
            WHERE status = ''accepted''
            AND tc.name = ''customerrequest''
            ORDER BY owner, CAST(p.value AS int), t.type, time'
            where id=4;""")
    cursor.execute("""UPDATE report set query='
            SELECT p.value AS __color__,
               owner AS __group__,
               t.id AS ticket, summary, milestone, t.type AS type, cr.name AS CR, time AS created,
               t.description AS _description_,
               changetime AS _changetime, reporter AS _reporter
            FROM ticket t
            LEFT JOIN enum p ON p.name = t.priority AND p.type = ''priority''
            LEFT JOIN ticket_custom tc ON t.id = tc.ticket
            LEFT JOIN public.customer_requests cr ON tc.value = cr.id
            WHERE status = ''accepted''
            AND tc.name = ''customerrequest''
            ORDER BY owner, CAST(p.value AS int), t.type, time'
            where id=5;""")
    cursor.execute("""UPDATE report set query='
            SELECT p.value AS __color__,
            (CASE status WHEN ''accepted'' THEN ''Accepted'' ELSE ''Owned'' END) AS __group__,
                 t.id AS ticket, summary, milestone, cr.name AS CR,
                 t.type AS type, priority, time AS created,
                 changetime AS _changetime, t.description AS _description,
                 reporter AS _reporter
            FROM ticket t
            LEFT JOIN enum p ON p.name = t.priority AND p.type = ''priority''
            LEFT JOIN ticket_custom tc ON t.id = tc.ticket
            LEFT JOIN public.customer_requests cr ON tc.value = cr.id
            WHERE t.status <> ''closed'' AND owner = $USER
            AND tc.name = ''customerrequest''
            ORDER BY (status = ''accepted'') DESC, CAST(p.value AS int), milestone, t.type, time'
            where id=7;""")
    cursor.execute("""UPDATE report set query='
            SELECT p.value AS __color__,
           (CASE owner
            WHEN $USER THEN ''My Tickets''
                ELSE ''Active Tickets''
            END) AS __group__,
                t.id AS ticket, summary, milestone, t.type AS type, cr.name AS CR,
                owner, status,
                time AS created,
                changetime AS _changetime, t.description AS _description,
                reporter AS _reporter
            FROM ticket t
            LEFT JOIN enum p ON p.name = t.priority AND p.type = ''priority''
            LEFT JOIN ticket_custom tc ON t.id = tc.ticket
            LEFT JOIN public.customer_requests cr ON tc.value = cr.id
            WHERE status <> ''closed''
            AND tc.name = ''customerrequest''
            ORDER BY (COALESCE(owner, '''') = $USER) DESC, CAST(p.value AS int), milestone, t.type, time'
            where id=8;""")
    cursor.close()
    cnx.commit()

    # remove default trac's milestones, components, versions
    cursor = cnx.cursor()
    cursor.execute("DELETE FROM milestone;")
    cursor.close()
    cnx.commit()

    cursor = cnx.cursor()
    cursor.execute("DELETE FROM component;")
    cursor.close()
    cnx.commit()

    cursor = cnx.cursor()
    cursor.execute("DELETE FROM version;")
    cursor.close()
    cnx.commit()


    # TODO: attualmente il riferimento dal trac al progetto dashboard è il project_id,
    #       considerando un'instalalzzione che condicide lo stesso stack wsgi,
    #       valutare se deve essere una uri (jsnorpc, xmlrpx, ...)
    #       o un dsn che punti al db, o ...
    tracenv.config.set('por-dashboard', 'project-id', application.project_id)

    # custom templates
    templates = settings.get('penelope.trac.templates')
    masterconfig = settings.get('penelope.trac.masterconfig')

    if templates:
        tracenv.config.set('inherit', 'templates_dir', templates)

    # master config
    if masterconfig:
        tracenv.config.set('inherit', 'file', masterconfig)

    # set name and description
    tracenv.config.set('project', 'name', project_name)
    tracenv.config.set('project', 'descr', application.name)

    tracenv.config.set('notification', 'smtp_enabled', smtp_enabled and 'true' or 'false')
    tracenv.config.set('notification', 'always_notify_owner', 'true')
    tracenv.config.set('notification', 'always_notify_reporter', 'true')
    manager_email = project.manager.email or ''
    tracenv.config.set('notification', 'smtp_always_cc', manager_email) # manager should always receiv CC

    tracenv.config.set('attachment', 'max_size', attachment_max_size)

    tracenv.config.set('components', 'penelope.trac.*', 'enabled')
    tracenv.config.set('components', 'themeengine.admin.*', 'enabled')
    tracenv.config.set('components', 'themeengine.api.*', 'enabled')
    tracenv.config.set('components', 'themeengine.web_ui.*', 'enabled')
    tracenv.config.set('components', 'ticketrelations.*', 'enabled')
    tracenv.config.set('components', 'tracopt.perm.config_perm_provider.extrapermissionsprovider', 'enabled')

    # All the custom permission names are converted to uppercase.
    # It is not possible to have lowercase and distinguish them from the standard permissions.
    tracenv.config.set('extra-permissions', 'extra_timeentry', 'TIME_ENTRY_ADD')

    tracenv.config.set('theme','theme', 'por')
    # ticket-custom
    tracenv.config.set('ticket-custom', 'customerrequest', 'select')
    tracenv.config.set('ticket-custom', 'customerrequest.label', 'Customer Request')
    tracenv.config.set('ticket-custom', 'blocking', 'text')
    tracenv.config.set('ticket-custom', 'blocking.label', 'Blocking')
    tracenv.config.set('ticket-custom', 'blockedby', 'text')
    tracenv.config.set('ticket-custom', 'blockedby.label', 'Blocked By')
    # BBB: ii valori di customerrequest vengono caricati ondemand
    tracenv.config.set('ticket-custom', 'customerrequest.options', '')

    # see ticket:80
    if qafields:
        tracenv.config.set('ticket-custom', 'issuetype', 'select')
        tracenv.config.set('ticket-custom', 'issuetype.label', 'Natura del problema')
        tracenv.config.set('ticket-custom', 'issuetype.options', '|'.join([u"",
                                u"sistemistica",
                                u"funzionalità",
                                u"design (grafica, layout...)",
                                u"prestazioni",
                                u"mi aspettavo che..."]))
        tracenv.config.set('ticket-custom', 'esogeno', 'checkbox')
        tracenv.config.set('ticket-custom', 'esogeno.label', 'Ticket aperto dal Cliente')
        tracenv.config.set('ticket-custom', 'esogeno.value', 'false')
        tracenv.config.set('ticket-custom', 'stats_exclude', 'checkbox')
        tracenv.config.set('ticket-custom', 'stats_exclude.label', 'Exclude from report stats')
        tracenv.config.set('ticket-custom', 'stats_exclude.value', 'false')
        tracenv.config.set('ticket-custom', 'fasesviluppo', 'select')
        tracenv.config.set('ticket-custom', 'fasesviluppo.label', 'Fase sviluppo')
        tracenv.config.set('ticket-custom', 'fasesviluppo.options', '|'.join([u"",
                                u"In lavorazione",
                                u"Per lo staging",
                                u"In staging",
                                u"Per la produzione",
                                u"In produzione"]))

    tracenv.config.set('ticket', 'restrict_owner', 'true')

    tracenv.config.set('ticket-custom', 'milestone.invalid_if', '')

    # WORKFLOW
    tracenv.config.set('ticket-workflow', 'accept', 'new,reviewing -> assigned')
    tracenv.config.set('ticket-workflow', 'accept.operations', 'set_owner_to_self')
    tracenv.config.set('ticket-workflow', 'accept.permissions', 'TICKET_MODIFY')
    tracenv.config.set('ticket-workflow', 'leave', '* -> *')
    tracenv.config.set('ticket-workflow', 'leave.default', '1')
    tracenv.config.set('ticket-workflow', 'leave.operations', 'leave_status')
    tracenv.config.set('ticket-workflow', 'reassign', 'new,assigned,accepted,reopened -> assigned')
    tracenv.config.set('ticket-workflow', 'reassign.operations', 'set_owner')
    tracenv.config.set('ticket-workflow', 'reassign.permissions', 'TICKET_MODIFY')
    tracenv.config.set('ticket-workflow', 'reassign_reviewing', 'reviewing -> *')
    tracenv.config.set('ticket-workflow', 'reassign_reviewing.name', 'reassign review')
    tracenv.config.set('ticket-workflow', 'reassign_reviewing.operations', 'set_owner')
    tracenv.config.set('ticket-workflow', 'reassign_reviewing.permissions', 'TICKET_MODIFY')
    tracenv.config.set('ticket-workflow', 'reopen', 'closed -> reopened')
    tracenv.config.set('ticket-workflow', 'reopen.operations', 'del_resolution')
    tracenv.config.set('ticket-workflow', 'reopen.permissions', 'TRAC_ADMIN')
    tracenv.config.set('ticket-workflow', 'resolve', 'new,assigned,reopened,reviewing -> closed')
    tracenv.config.set('ticket-workflow', 'resolve.operations', 'set_resolution')
    tracenv.config.set('ticket-workflow', 'resolve.permissions', 'TICKET_MODIFY')
    tracenv.config.set('ticket-workflow', 'review', 'new,assigned,reopened -> reviewing')
    tracenv.config.set('ticket-workflow', 'review.operations', 'set_owner')
    tracenv.config.set('ticket-workflow', 'review.permissions', 'TICKET_MODIFY')

    tracenv.config.set('milestone-groups', 'closed', 'closed')
    tracenv.config.set('milestone-groups', 'closed.order', '0')
    tracenv.config.set('milestone-groups', 'closed.query_args', 'group=resolution')
    tracenv.config.set('milestone-groups', 'closed.overall_completion', 'true')

    tracenv.config.set('milestone-groups', 'active', '*')
    tracenv.config.set('milestone-groups', 'active.order', '1')
    tracenv.config.set('milestone-groups', 'active.css_class', 'open')

    tracenv.config.set('milestone-groups', 'new', 'new,reopened')
    tracenv.config.set('milestone-groups', 'new.order', '2')


    # navigation
    tracenv.config.set('metanav', 'logout', 'disabled')

    # permissions
    tracenv.config.set('components', 'penelope.trac.users.*', 'enabled')
    tracenv.config.set('trac', 'permission_store', 'PorPermissionStore')
    tracenv.config.set('trac', 'show_email_addresses', 'true')

    # xmlrpc plugin
    tracenv.config.set('components', 'tracrpc.api.xmlrpcsystem', 'enabled')
    tracenv.config.set('components', 'tracrpc.xml_rpc.xmlrpcprotocol', 'enabled')
    tracenv.config.set('components', 'tracrpc.web_ui.rpcweb', 'enabled')
    tracenv.config.set('components', 'tracrpc.ticket.*', 'enabled')

    # DynamicFields Plugin
    tracenv.config.set('components', 'dynfields.rules.*','enabled')
    tracenv.config.set('components', 'dynfields.web_ui.*','enabled')

    # User & Roles (ReadOnly !!!)
    # tracenv.config.set('user_manager', 'user_store', 'PorUserStore')
    # tracenv.config.set('user_manager', 'attribute_provider', 'PorAttributeProvider')

    # BatchModifyPlugin
    if batchmod:
        tracenv.config.set('components', 'batchmod.web_ui.*', 'enabled')

    # Traccustomfieldadmin
    if customfieldadmin:
        tracenv.config.set('components', 'customfieldadmin.*', 'enabled')

    # PrivateCommentsPlugin
    if privatecomments:
        tracenv.config.set('components', 'privatecomments.privatecomments.*', 'enabled')

    # PrivateTicketPlugin
    if privatetickets:
        tracenv.config.set('components', 'privatetickets.policy.*', 'enabled')
        tracenv.config.set('trac', 'permission_policies',
            'PrivateTicketsPolicy, %s' % tracenv.config.get('trac', 'permission_policies'))
        trac_perm['customer'].append('TICKET_VIEW_SELF')

    # SensitiveTicketsPlugin
    if sensitivetickets:
        tracenv.config.set('components', 'sensitivetickets.*', 'enabled')
        tracenv.config.set('trac', 'permission_policies',
            'SensitiveTicketsPolicy, %s' % tracenv.config.get('trac', 'permission_policies'))
        tracenv.config.set('ticket-custom', 'sensitive.show_if_group', 'administrator|developer')
        # utilizzato se il default e' 1, se il default e' 0 non serve
        # tracenv.config.set('ticket-custom', 'sensitive.clear_on_hide', 'false')
        # tracenv.config.set('ticket-custom', 'sensitive.has_permission', 'SENSITIVE_VIEW')
        tracenv.config.set('ticket-custom', 'sensitive.value', '0')
        trac_perm['developer'].append('SENSITIVE_VIEW')

    # tracwysiwyg
    if tracwysiwyg:
        tracenv.config.set('components', 'tracwysiwyg.wysiwygmodule', 'enabled')

    # AutoCompleteUsers
    if autocompleteusers:
        tracenv.config.set('components', 'autocompleteusers.autocompleteusers', 'enabled')

    tracenv.config.set('wiki', 'max_size', 1048576)

    tracenv.config.set('logging', 'log_file', 'trac.log')
    tracenv.config.set('logging', 'log_level', 'INFO')
    tracenv.config.set('logging', 'log_type', 'file')

    tracenv.config.save()

    # let's remove notification config - it is set by the global config
    tracenv.config.remove('notification', 'smtp_from')
    tracenv.config.remove('notification', 'smtp_from_name')
    tracenv.config.remove('notification', 'replyto')
    tracenv.config.remove('notification', 'smtp_replyto')
    tracenv.config.remove('notification', 'email_sender')
    tracenv.config.remove('notification', 'smtp_enabled')
    tracenv.config.remove('notification', 'smtp_host')
    tracenv.config.remove('notification', 'smtp_port')
    tracenv.config.remove('notification', 'smtp_password')
    tracenv.config.remove('notification', 'smtp_username')

    tracenv.config.remove('trac', 'repository_sync_per_request')
    tracenv.config.save()

    run([trac_path, 'upgrade --no-backup'])

    run([trac_path, 'permission remove anonymous *'])
    run([trac_path, 'permission remove authenticated *'])
    for role, perms in trac_perm.items():
        for perm in perms:
            run([trac_path, "permission add %s %s" % (role, perm)])

    run([trac_path, "ticket_type add verification"])

    # hack to minimize config size
    run([trac_path, 'config set browser color_scale True'])

    # add properly milestones
    milestones.append({'title': 'Backlog', 'due_date': date.today().replace(year=date.today().year+1)})

    for milestone in milestones:
        due = milestone['due_date']
        if due:
            due = milestone['due_date'].strftime('%Y-%m-%d')
            run([trac_path, 'milestone add "%s" %s' % (milestone['title'], due)])
        else:
            run([trac_path, 'milestone add "%s"' % milestone['title']])

    tracenv = Environment(trac_path)
    for ticket in tickets:
        # in this moment the customer request has a proper id
        ticket['status'] = 'new'
        t = Ticket(tracenv)
        t.populate(ticket)
        t.insert()

    application.api_uri = 'trac://%s' % tracname
    application.trac_name = tracname

# !!!!!!!!!
# !!!!!!!!!
# ADD RANDOM TICKETS
def add_trac_tickets(project, customer_request, settings, users):
    g = lipsum.Generator()
    tracname = idnormalizer.normalize(project.name)
    tracenvs = settings.get('penelope.trac.envs')
    trac_path = '%s/%s' % (tracenvs, tracname)
    tracenv = Environment(trac_path)
    for i in range(50):
        ticket = Ticket(tracenv)
        ticket.populate({
            'summary': g.generate_sentence(),
            'description': g.generate_paragraph(),
            # 'project_id': project.id,
            # 'customerrequest': "%s>%s>%s" % (project.id, customer_request.id, customer_request.name),
            'owner': random.choice(users).email,
            'customerrequest': "%s" % (customer_request.id),
            'status': 'new',
        })
        ticket.insert()
        yield ticket





from trac import core
from trac.perm import IPermissionStore, DefaultPermissionStore, IPermissionGroupProvider

from penelope.core.models import DBSession
from penelope.core.models.dashboard import Project, User

 
class PorPermissionStore(DefaultPermissionStore):
    """ """
    core.implements(IPermissionStore, IPermissionGroupProvider)

    # IPermissionGroupProvider
    def get_permission_groups(self, username):
        # TODO: work only for por/trac on the same wsgi stack
        project_id = self.env.config.get('por-dashboard', 'project-id')
        if project_id:
            project = DBSession().query(Project).get(project_id)
            user = DBSession().query(User).filter_by(email=username).first()
            if user:
                return list(user.roles_in_context(context=project))
        return list() 

    # IPermissionStore
    def get_user_permissions(self, username):
        """Return all permissions for the user with the specified name.

        The permissions are returned as a dictionary where the key is the name
        of the permission, and the value is either `True` for granted
        permissions or `False` for explicitly denied permissions."""

        # TODO: work only for por/trac on the same wsgi stack
        actions = set(super(PorPermissionStore, self).get_user_permissions(username))
        project_id = self.env.config.get('por-dashboard', 'project-id')
        if project_id:
            project = DBSession().query(Project).get(project_id)
            user = DBSession().query(User).filter_by(email=username).first()
            if user:
                for role in user.roles_in_context(context=project):
                    actions.update(set(super(PorPermissionStore, self).get_user_permissions(role)))
        return list(actions) 



# -*- coding: utf-8 -*-
import logging

from penelope.core.models import DBSession
from penelope.core.lib.helpers import unicodelower

log = logging.getLogger(__name__)


def is_inside_penelope():
    """
    This is a way to sense whether the components
    are run inside a web application or in trac-admin.
    """
    if DBSession.bind:
        return True
    return False


def fix_connection_init():
    """
    patch per gestire postgresql 9.x dve l'eccezione nella gestione degli
    schema e' cambiata (TODO: segnalare sul trac di trac.edgewall.org)
    """

    import psycopg2
    from psycopg2 import DataError, ProgrammingError as PGSchemaError
    from trac.db.util import ConnectionWrapper
    from trac.db.postgres_backend import assemble_pg_dsn, PostgreSQLConnection

    # l'eccezione ProgrammingError cambia in PGSchemaError
    def PostgreSQLConnection__init__(self, path, log=None, user=None, password=None, host=None,
                port=None, params={}):
        if path.startswith('/'):
            path = path[1:]
        if 'host' in params:
            host = params['host']

        cnx = psycopg2.connect(assemble_pg_dsn(path, user, password, host, port))

        cnx.set_client_encoding('UNICODE')
        try:
            self.schema = None
            if 'schema' in params:
                self.schema = params['schema']
                cnx.cursor().execute('SET search_path TO %s', (self.schema,))
                cnx.commit()
        # except (DataError, ProgrammingError):
        except PGSchemaError:
            cnx.rollback()
        except DataError:
            cnx.rollback()
        ConnectionWrapper.__init__(self, cnx, log)

    PostgreSQLConnection.__init__ = PostgreSQLConnection__init__


def fix_get_custom_fields():
    """
    patch per custom field con dati su por
    """

    import copy
    from trac.ticket.api import TicketSystem
    from penelope.core.models.dashboard import Project

    # TODO: cache ?
    # TODO: generalizzare
    def TicketSystem_get_custom_fields(self):
        if not is_inside_penelope():        # we are in trac-admin
            return self.custom_fields

        custom_fields = copy.deepcopy(self.custom_fields)
        project_id = self.config.get('por-dashboard', 'project-id')
        project = None
        if project_id:
            project = DBSession().query(Project).get(project_id)
        for field in custom_fields:
            if project and field['name'] == 'customerrequest':
                field['options'] = [cr.id for cr in sorted(project.customer_requests, key=unicodelower)]
                field['descriptions'] = dict([(cr.id, cr.name) for cr in project.customer_requests])
        # TODO: rimuovere il metodo sotto che invalida la cache, trovare
        # un altro modo per mantenere aggiornati i dati
        self.reset_ticket_fields()
        return custom_fields

    TicketSystem.get_custom_fields = TicketSystem_get_custom_fields


def fix_send_user_error():
    """
    patch per 401/403
    """

    import sys
    from genshi.builder import Fragment, tag
    from trac.util.text import exception_to_unicode
    from trac.web.api import RequestDone
    from trac.util.text import to_unicode
    from trac.util.translation import _, tag_
    import trac.web.main

    def send_user_error(req, env, e):
        # See trac/web/api.py for the definition of HTTPException subclasses.
        if env:
            env.log.warn('[%s] %s' % (req.remote_addr, exception_to_unicode(e)))
        try:
            # We first try to get localized error messages here, but we
            # should ignore secondary errors if the main error was also
            # due to i18n issues
            title = _("Error")
            if e.reason:
                if title.lower() in e.reason.lower():
                    title = e.reason
                else:
                    title = _("Error: %(message)s", message=e.reason)
        except Exception:
            title = 'Error'
        # The message is based on the e.detail, which can be an Exception
        # object, but not a TracError one: when creating HTTPException,
        # a TracError.message is directly assigned to e.detail
        if isinstance(e.detail, Exception): # not a TracError
            message = exception_to_unicode(e.detail)
        elif isinstance(e.detail, Fragment): # markup coming from a TracError
            message = e.detail
        else:
            message = to_unicode(e.detail)
        data = {'title': title, 'type': 'TracError', 'message': message,
                'frames': [], 'traceback': None}
        if e.code == 403 and req.authname == 'anonymous':
            # TRANSLATOR: ... not logged in, you may want to 'do so' now (link)
            do_so = tag.a(_("do so"), href=req.href.login())
            req.chrome['notices'].append(
                tag_("You are currently not logged in. You may want to "
                    "%(do_so)s now.", do_so=do_so))
            # MONKEY PATCH HERE !!!
            e.code = 401
        try:
            req.send_error(sys.exc_info(), status=e.code, env=env, data=data)
        except RequestDone:
            pass

    trac.web.main._send_user_error = send_user_error


def fix_get_known_users():
    """
    patch per known_user su por
    """

    from trac.env import Environment
    from penelope.core.models.dashboard import Project, User

    # TODO: cache ?
    # TODO: esistono api piu' semplici su por per la stessa richiesta?
    def Environment_get_known_users(self, cnx=None):
        project_id = self.config.get('por-dashboard', 'project-id')
        project = None
        if project_id:
            db = DBSession()
            project = db.query(Project).get(project_id)
            for user in db.query(User).all():
                if user.roles_in_context(project):
                    yield user.login, user.fullname, user.email

    Environment.get_known_users = Environment_get_known_users


def fix_customer_request_changelog_description():
    """
    Look up Customer Request description while creating change summaries (both web and email)
    """

    from trac.ticket.web_ui import TicketModule
    from penelope.core.models.dashboard import CustomerRequest

    _grouped_changelog_entries = TicketModule.grouped_changelog_entries
    def TicketModule_grouped_changelog_entries(self, ticket, db, when=None):
        ret = _grouped_changelog_entries(self, ticket, db, when)
        for item in ret:
            try:
                cr = item['fields']['customerrequest']
                qry = DBSession().query(CustomerRequest)
                old_cr = qry.get(cr['old'])
                new_cr = qry.get(cr['new'])
                cr['old'] = old_cr.name if old_cr else cr['old']
                cr['new'] = new_cr.name if new_cr else cr['new']
            except KeyError:
                pass

            yield item

    TicketModule.grouped_changelog_entries = TicketModule_grouped_changelog_entries


def fix_customer_request_dropdown():
    """
    Renders the CR dropdown options grouping them by state
    """

    from trac.ticket.web_ui import TicketModule
    from penelope.core.models.dashboard import CustomerRequest

    cr_order = ['estimated', 'created','scheduled', 'achieved', 'invoiced']

    def cr_sortkey(cr):
        try:
            return cr_order.index(cr.workflow_state)
        except (AttributeError, ValueError):
            return -1

    contract_order = ['active', 'draft','done']

    def contract_sortkey(group):
        try:
            return contract_order.index(group['label'].workflow_state)
        except (AttributeError, ValueError):
            return -1

    def prepare_customerrequest_options(field, newticket):
        qry = DBSession.query(CustomerRequest)
        options = field['options']
        customer_requests = [qry.get(op) for op in options]
        if newticket:
            customer_requests = [cr for cr in customer_requests if cr.workflow_state in ['created', 'estimated']]

        groups = {}
        NO_CONTRACT = 'No contract available'
        for cr in sorted(customer_requests, key=cr_sortkey):
            if cr is None: # the CR has probably been deleted
                continue
            contract = cr.contract and cr.contract or NO_CONTRACT
            groups.setdefault(contract, {
                                    'label': contract,
                                    'options': [],
                                    'descriptions': [],
                                })
            groups[contract]['options'].append(cr.id)
            groups[contract]['descriptions'].append(cr.name)
        field['options'] = []
        field['descriptions'] = []
        field['optgroups'] = sorted(groups.values(), key=contract_sortkey)
        field['optional'] = True

    _prepare_fields = TicketModule._prepare_fields
    def TicketModule_prepare_fields(self, req, ticket):
        ret = _prepare_fields(self, req, ticket)
        for field in ret:
            if field['name'] == 'customerrequest':
                if not ticket.id:
                    newticket = True
                else:
                    newticket = False
                prepare_customerrequest_options(field, newticket)

        return ret

    TicketModule._prepare_fields = TicketModule_prepare_fields


def fix_filter_email_recipents():
    """
    FIX: per evitare mail al customer relativamente a commenti privati e ticket sensibili
    """
    from trac.ticket.notification import TicketNotifyEmail
    from trac.ticket.web_ui import TicketModule
    from trac.perm import PermissionSystem
    try:
        import privatecomments; privatecomments
        HAS_PRIVATECOMMENTS = True
    except ImportError:
        HAS_PRIVATECOMMENTS = False


    TicketNotifyEmail._orig_get_recipients = TicketNotifyEmail.get_recipients

    def TicketNotifyEmail_get_recipients(self, tktid):
        (torecipients, ccrecipients) = self._orig_get_recipients(tktid)
        perm = PermissionSystem(self.env)
        # sensitivetickets
        if self.ticket['sensitive'] == '1':
            def has_sensisitive_perm(username):
                return perm.get_user_permissions(username).get('SENSITIVE_VIEW')
            torecipients = filter(has_sensisitive_perm, torecipients)
            ccrecipients = filter(has_sensisitive_perm, ccrecipients)
        # privatecomments
        if HAS_PRIVATECOMMENTS:
            privatecomment = False
            for mod in TicketModule(self.env).grouped_changelog_entries(self.ticket, self.db, self.modtime):
                cursor = self.db.cursor()
                sql = 'SELECT private FROM private_comment WHERE ticket_id=%d AND comment_id=%d AND private>0'
                cursor.execute(sql % (int(self.ticket.id), int(mod.get('cnum'))))
                if cursor.cursor.rowcount:
                    privatecomment = True
                cursor.close()
            if privatecomment:
                def has_privatecommente_perm(username):
                    return perm.get_user_permissions(username).get('PRIVATE_COMMENT_PERMISSION')
                torecipients = filter(has_privatecommente_perm, torecipients)
                ccrecipients = filter(has_privatecommente_perm, ccrecipients)
        return (torecipients, ccrecipients)

    TicketNotifyEmail.get_recipients = TicketNotifyEmail_get_recipients


def fix_notification_props():
    """
    Change the table of ticket properties printed in notifications.
    """
    from trac.ticket.notification import TicketNotifyEmail

    def iter_props(self):
        tkt = self.ticket
        yield '--'

        for f in tkt.fields:
            fname = f['name']

            if fname in ['summary', 'cc', 'time', 'changetime', 'sensitive', 'esogeno', 'customerrequest', 'stats_exclude']:
                continue

            fval = tkt[fname] or ''
            if fname in ['owner', 'reporter']:
                fval = self.obfuscate_email(fval)

            flabel = f['label']

            if not fval:
                continue

            if f['type'] == 'textarea':
                yield u'%s:' % flabel
                for line in fval.split('\n'):
                    yield u'      %s' % line
            else:
                yield u'%s: %s' % (flabel, fval)


    def TicketNotifyEmail_format_props(self):
        return '\n'.join(iter_props(self))

    TicketNotifyEmail.format_props = TicketNotifyEmail_format_props


log.info("Monkey patch")
fix_connection_init()
fix_get_custom_fields()
fix_send_user_error()
fix_get_known_users()
fix_customer_request_changelog_description()
fix_customer_request_dropdown()
fix_filter_email_recipents()
fix_notification_props()



# -*- coding: utf-8 -*-
import hashlib
import os
import transaction

from ConfigParser import ConfigParser
from sqlalchemy.orm.exc import NoResultFound
from time import time

from penelope.core.models import DBSession, includeme
from penelope.core.models.dashboard import User


# TODO: sostituire con un config parser vero?
class Config(object):
    ''' fake config parser ... '''
    def __init__(self, ini):
        self.cfg = ConfigParser()
        self.cfg.read(ini)

    @property
    def registry(self):
        class Registry:
            def __init__(self, cfg):
                self.cfg = cfg
            @property
            def settings(self):
                return dict(self.cfg.items('app:dashboard'))
        return Registry(self.cfg)

# TODO: cache

def check_password(environ, login, password):
    hash = hashlib.md5('%s:%s' % (login, password)).hexdigest()
    if int(time()) - cache.get(hash, 0) < TIMEOUT:
        return True

    db = DBSession()
    try:
        try:
            user = db.query(User).filter_by(svn_login=login).one()
        except NoResultFound:
            return None
        if user.check_password(password):
            cache[hash] = int(time())
            return True
        else:
            return False
    finally:
        transaction.commit()

TIMEOUT=30
cache = {}
includeme(Config(os.environ['POR_INI']))

def main(*args):
    pass



#!/usr/bin/env py
"""
This script modifies the Trac CSS files to restrict their scope inside a specific container.

It makes use of http://packages.python.org/cssutils/README.html

for testing, you can do
csstidy trac-orig.css --preserve_css=true --template=low --silent=true | sed s/\\t/\ \ /g > trac-tidy.css
"""

import os
import pkg_resources
import sys

import cssutils

container_id = '#trac-container'


def new_part(part):
    words = part.split()
    if any(w.startswith('body') for w in words):
        return ' '.join('body %s' % container_id if p=='body' else p for p in words)
    elif any(w.startswith('html') for w in words):
        return ' '.join('html %s' % container_id if p=='html' else p for p in words)
    return '%s %s' % (container_id, part)



def update_rule(rule):
    rule.selectorText = ','.join(new_part(p) for p in rule.selectorText.split(','))
    for css_property in rule.style:
        css_property.value = css_property.value.replace('../', '../../common/')


def process_css(body):
    cssutils.ser.prefs.indent = ' '*2
    cssutils.ser.prefs.indentClosingBrace = False
    cssutils.ser.prefs.listItemSpacer = ''
    cssutils.ser.prefs.omitLastSemicolon = False
    cssutils.ser.prefs.omitLeadingZero = True
    cssutils.ser.prefs.propertyNameSpacer = ''

    sheet = cssutils.parseString(body)
    for rule in sheet:
        if isinstance(rule, cssutils.css.CSSStyleRule):
            update_rule(rule)
        elif isinstance(rule, cssutils.css.cssmediarule.CSSMediaRule):
            for subrule in rule:
                if isinstance(subrule, cssutils.css.CSSStyleRule):
                    update_rule(subrule)

    output = sheet.cssText
    output = output.replace(' {', '\n{')
    output = output.replace('}', '}\n')
    return output.rstrip('\n')




def run(dirpath):
    if not os.path.isdir(dirpath):
        raise ValueError('Not a directory: %s' % dirpath)

    # code.css first, as it is included by others
    allfiles = sorted(os.listdir(dirpath), key=lambda x: (not x.startswith('code.'), x))

    target_dir = pkg_resources.resource_filename('penelope.trac', 'htdocs/rescoped-css')
    os.chdir(target_dir)

    for fname in allfiles:
        with open(os.path.join(dirpath, fname)) as fin:
            print fname
            out = process_css(body=fin.read())
            with open(os.path.join(target_dir, fname), 'w') as fout:
                fout.write('/* This file has been automatically generated by trac-rescope. Do not edit. */\n')
                fout.write(out)
    print 'done.'



if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('Usage: %s [egg | /path/to/trac/htdocs/css]' % sys.argv[0])
        sys.exit(10)

    if sys.argv[1] == 'egg':
        htdocs = pkg_resources.resource_filename('trac', 'htdocs')
        dirpath = os.path.join(htdocs, 'css')
    else:
        dirpath = sys.argv[1]
    print 'Processing CSS from %s...' % dirpath
    run(dirpath=dirpath)






# -*- coding: utf-8 -*-
#
# http://trac-hacks.org/wiki/TicketValidatorPlugin
#

from trac.core import Component, implements
from trac.ticket.api import TicketSystem
from trac.ticket import ITicketManipulator


class TicketWorkflowGuards(Component):
    """ """

    implements(ITicketManipulator)

    def prepare_ticket(self, req, ticket, fields, actions):
        """Not currently called, but should be provided for future
           compatibility."""

    def validate_ticket(self, req, ticket):
        """Make sure required fields for the next state have been
           the ticket will be in have been entered."""

        state = self._get_state(req, ticket)

        errors = []

        if ticket['milestone'] == '':
            errors.append(('milestone', 'Milestone cannot be empty'))

        if ticket['type'] == 'defect' and ticket['issuetype'] == '':
            errors.append(('issuetype', 'Devi specificare la natura del problema se si tratta di un difetto'))

        return errors

    def _get_state(self, req, ticket):
        """Get the state this ticket is going to be in."""

        if 'action' not in req.args:
            return 'new'

        action = req.args['action']
        action_changes = {}

        for controller in self._get_action_controllers(req, ticket, action):
            action_changes.update(controller.get_ticket_changes(req, ticket, action))

        return 'status' in action_changes and action_changes['status'] or ticket['status']

    def _get_action_controllers(self, req, ticket, action):
        for controller in TicketSystem(self.env).action_controllers:
            actions = [action for weight, action in
                       controller.get_ticket_actions(req, ticket)]
            if action in actions:
                yield controller



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = """penelope.trac
=============

for more details visit: http://getpenelope.github.com/"""
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()

requires = [
    'distribute',
    'mandrill',
    'penelope.core',
    'privatecomments',
    'python-creole',
    'sensitivetickets>=0.21',
    'setuptools',
    'Trac',
    'TracDynamicFields',
    'TracThemeEngine>=2.0',
    'TracXMLRPC',
    ]

extra = {} 
try:
    from trac.util.dist import get_l10n_cmdclass 
    cmdclass = get_l10n_cmdclass() 
    if cmdclass: # Yay, Babel is there, we've got something to do! 
        extra['cmdclass'] = cmdclass 
        extractors = [ 
            ('**.py',                'python', None), 
            ('**/templates/**.html', 'genshi', None), 
            ('**/templates/**.txt',  'genshi', { 
                'template_class': 'genshi.template:TextTemplate' 
            }), 
        ] 
        extra['message_extractors'] = { 
            'penelope/trac': extractors, 
        }
except ImportError: 
    pass 


setup(name='penelope.trac',
      version='1.2.12',
      description='Penelope: Trac integration',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Pylons",
        ],
      author='Penelope Team',
      author_email='penelopedev@redturtle.it',
      url='http://getpenelope.github.com',
      keywords='web wsgi bfg pylons pyramid',
      namespace_packages=['penelope'],
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='penelope.trac',
      install_requires = requires,
      entry_points = {
          'console_scripts': [
            'auth_wsgi = penelope.trac.auth_wsgi:main',
          ],
          'trac.plugins': [
            'penelope.trac = penelope.trac.plugins',
            'penelope.trac.users = penelope.trac.user',
            'penelope.trac.communication = penelope.trac.communication',
            'penelope.trac.workflow = penelope.trac.workflow',
          ]
      },
      **extra
      )
